@@ -3,20 +3,31 @@ A collection of functions for integrating with the Archimedes API
 """
 
 import json
+import os
+from functools import partial
 from http import HTTPStatus
 from typing import Dict, List
 
 import pandas as pd
 import requests
-from requests.exceptions import HTTPError, JSONDecodeError
+from requests.exceptions import (
+    ConnectionError,
+    ConnectTimeout,
+    HTTPError,
+    JSONDecodeError,
+    Timeout,
+)
+from retry import retry
 
 import archimedes
-from archimedes.auth import get_auth, NoneAuth
-from archimedes.configuration import get_api_config
-from archimedes.data.types import Prediction, PredictionData
+from archimedes.auth import NoneAuth, get_auth
+from archimedes.configuration import get_api_base_url
+from archimedes.data.types import PredictionData
 
-API_URL = get_api_config().url
+DEFAULT_TIMEOUT = os.getenv("ARCHIMEDES_API_TIMEOUT", 120)  # 2 minutes
 API_VERSION = 2
+get_api_base_url_v2 = partial(get_api_base_url, API_VERSION)
+RETRY_EXCEPTIONS = (ConnectionError, ConnectTimeout, HTTPError, Timeout)
 
 
 def get_intraday_trades(
@@ -74,19 +85,31 @@ def get_intraday_trades(
     else:
         end = pd.to_datetime(end)
 
-    query = {
-        "series_id": "NP/IntradayTrades",
-        "price_areas": price_areas,
-        "start": start,
-        "end": end,
-        "flatten_columns": False,
-    }
+    queries = [
+        {
+            "series_id": "NP/IntradayTrades",
+            "start": start,
+            "end": end,
+            "flatten_columns": False,
+        }
+    ]
+    if price_areas is not None:
+        queries = [
+            {"price_areas": price_area, **query}
+            for query in queries
+            for price_area in price_areas
+        ]
 
-    observation_data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/observation_json/get",
-        access_token=access_token,
-        params=query,
-    )
+    base_url = get_api_base_url_v2()
+    observation_data = []
+    for query in queries:
+        observation_data.extend(
+            _make_api_request(
+                f"{base_url}/observation_json/get",
+                access_token=access_token,
+                params=query,
+            )
+        )
 
     if not len(observation_data):
         return pd.DataFrame(
@@ -194,76 +217,119 @@ def get(
     else:
         end = pd.to_datetime(end)
 
-    query = {
-        "series_ids": series_ids,
-        "price_areas": price_areas,
-        "start": start,
-        "end": end,
-        "flatten_columns": True,
-    }
+    queries = [
+        {
+            "start": start,
+            "end": end,
+            "flatten_columns": True,
+        }
+    ]
+    if series_ids is not None and len(series_ids):
+        queries = [
+            {"series_ids": series_id, "series_id": series_id, **query}
+            for query in queries
+            for series_id in series_ids
+        ]
+    if price_areas is not None and len(price_areas):
+        queries = [
+            {"price_areas": price_area, **query}
+            for query in queries
+            for price_area in price_areas
+        ]
 
-    df = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/get",
-        access_token=access_token,
-        params=query,
-    )
-
-    if len(df):
-        df = pd.DataFrame.from_dict(df)
-        df = df.sort_values(by=["from_dt", "version"])
-    else:
-        df = pd.DataFrame(
-            columns=["series_id", "from_dt", "ref_dt", "value", "version", "price_area"]
-        )
-    df = df.pivot_table(
-        values="value",
-        columns=["series_id", "price_area"],
-        index="from_dt",
-        aggfunc="last",
-    )
-
-    del query["series_ids"]
-    for series_id in series_ids:
-        query["series_id"] = series_id
-        observation_data = _make_api_request(
-            f"{API_URL}/v{API_VERSION}/observation_json/get",
+    df = None
+    for query in queries:
+        base_url = get_api_base_url_v2()
+        observations = _make_api_request(
+            f"{base_url}/data/get",
             access_token=access_token,
             params=query,
         )
-        if len(observation_data):
-            observation_data = pd.DataFrame.from_dict(observation_data)
-
-            observation_data[
-                ["series_id1", "series_id2", "price_area"]
-            ] = observation_data["series_id"].str.split("/", 2, expand=True)
-            observation_data["series_id"] = (
-                observation_data["series_id1"] + "/" + observation_data["series_id2"]
-            )
-            observation_data = observation_data.drop(
-                ["series_id1", "series_id2"], axis=1
-            )
-
-            if series_id == "NP/IntradayTrades":
-
-                def agg_func(x):
-                    return list(x)
-
-                sort_headers = ["from_dt"]
-            else:
-                agg_func = "last"
-                sort_headers = ["from_dt", "version"]
-
-            observation_data = observation_data.sort_values(by=sort_headers)
-            observation_data = observation_data.pivot_table(
+        if len(observations):
+            observations = pd.DataFrame.from_dict(observations)
+            observations = observations.sort_values(by=["from_dt", "version"])
+            observations = observations.pivot_table(
                 values="value",
                 columns=["series_id", "price_area"],
                 index="from_dt",
-                aggfunc=agg_func,
+                aggfunc="last",
+            )
+        else:
+            observations = pd.DataFrame()
+
+        observation_json = _make_api_request(
+            f"{base_url}/observation_json/get",
+            access_token=access_token,
+            params=query,
+        )
+        if len(observation_json):
+            observation_json = pd.DataFrame.from_dict(observation_json)
+
+            observation_json[
+                ["series_id1", "series_id2", "price_area"]
+            ] = observation_json["series_id"].str.split("/", 2, expand=True)
+            observation_json["series_id"] = (
+                observation_json["series_id1"] + "/" + observation_json["series_id2"]
+            )
+            observation_json = observation_json.drop(
+                ["series_id1", "series_id2"], axis=1
+            )
+            is_intraday_trades = observation_json["series_id"] == "NP/IntradayTrades"
+
+            def agg_func(x):
+                return list(x)
+
+            observation_data_intraday = observation_json[is_intraday_trades]
+            if len(observation_data_intraday):
+                observation_data_intraday = observation_data_intraday.sort_values(
+                    by=["from_dt"]
+                )
+                observation_data_intraday = observation_data_intraday.pivot_table(
+                    values="value",
+                    columns=["series_id", "price_area"],
+                    index="from_dt",
+                    aggfunc=agg_func,
+                )
+            else:
+                observation_data_intraday = pd.DataFrame()
+            observation_json = observation_json[~is_intraday_trades]
+            if len(observation_json):
+                observation_json = observation_json.sort_values(
+                    by=["from_dt", "version"]
+                ).pivot_table(
+                    values="value",
+                    columns=["series_id", "price_area"],
+                    index="from_dt",
+                    aggfunc="last",
+                )
+            else:
+                observation_json = pd.DataFrame()
+
+            observation_json = observation_json.merge(
+                observation_data_intraday,
+                left_index=True,
+                right_index=True,
+                how="outer",
             )
 
-            df = pd.merge(
-                df, observation_data, left_index=True, right_index=True, how="outer"
+            observations = pd.merge(
+                observations,
+                observation_json,
+                left_index=True,
+                right_index=True,
+                how="outer",
             )
+        df = (
+            observations
+            if df is None
+            else pd.merge(
+                df,
+                observations,
+                left_index=True,
+                right_index=True,
+                how="outer",
+            )
+        )
 
     df.index = pd.to_datetime(df.index)
     return df
@@ -386,23 +452,35 @@ def get_predictions(
     if ref_dt_end is not None:
         query["ref_dt_end"] = pd.to_datetime(ref_dt_end, utc=True)
 
+    queries = [query]
     if series_ids is not None:
-        query["series_ids"] = series_ids
+        queries = [
+            {"series_ids": series_id, **query}
+            for query in queries
+            for series_id in series_ids
+        ]
 
     if price_areas is not None:
-        query["price_areas"] = price_areas
-
-    data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/get_predictions",
-        access_token=access_token,
-        params=query,
-    )
+        queries = [
+            {"price_areas": price_area, **query}
+            for query in queries
+            for price_area in price_areas
+        ]
 
+    base_url = get_api_base_url_v2()
+    data = []
+    for query in queries:
+        data.extend(
+            _make_api_request(
+                f"{base_url}/data/get_predictions",
+                access_token=access_token,
+                params=query,
+            )
+        )
     for item in data:
         json_data = item.get("json_data")
         item.update(json_data)
         del item["json_data"]
-
     df = pd.DataFrame.from_dict(data)
 
     date_fields = ["from_dt", "run_dt", "ref_dt"]
@@ -435,15 +513,16 @@ def list_series_price_areas(series_id: str, *, access_token: str = None):
     query = {
         "series_id": series_id,
     }
+    base_url = get_api_base_url_v2()
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas",
+        f"{base_url}/data/list_series_price_areas",
         access_token=access_token,
         params=query,
     )
     data = pd.DataFrame.from_dict(data)
 
     observation_data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/observation_json/list_series_price_areas",
+        f"{base_url}/observation_json/list_series_price_areas",
         access_token=access_token,
         params=query,
     )
@@ -473,13 +552,13 @@ def list_ids(sort: bool = False, *, access_token: str = None):
     Returns:
         DataFrame with all available list_ids
     """
-    data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_ids", access_token=access_token
-    )
+    base_url = get_api_base_url_v2()
+    data = _make_api_request(f"{base_url}/data/list_ids", access_token=access_token)
     data = pd.DataFrame.from_dict(data)
 
     observation_data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/observation_json/list_ids", access_token=access_token
+        f"{base_url}/observation_json/list_ids",
+        access_token=access_token,
     )
     observation_data = pd.DataFrame.from_dict(observation_data)
 
@@ -512,7 +591,8 @@ def list_prediction_ids(*, access_token: str = None):
     """
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_prediction_ids", access_token=access_token
+        f"{get_api_base_url_v2()}/data/list_prediction_ids",
+        access_token=access_token,
     )
 
     return pd.DataFrame.from_dict(data)
@@ -537,7 +617,7 @@ def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = No
         query["prediction_id"] = prediction_id
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/get_predictions_ref_dts",
+        f"{get_api_base_url_v2()}/data/get_predictions_ref_dts",
         access_token=access_token,
         params=query,
     )
@@ -545,6 +625,7 @@ def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = No
     return pd.DataFrame.from_dict(data)
 
 
+@retry(RETRY_EXCEPTIONS, tries=3, delay=2)
 def _make_api_request(url, method="GET", access_token=None, *args, **kwargs):
     if access_token is None:
         archimedes_auth = get_auth()
@@ -554,6 +635,11 @@ def _make_api_request(url, method="GET", access_token=None, *args, **kwargs):
             )
         access_token = archimedes_auth.get_access_token_silent()
 
+    timeout = kwargs.pop("timeout", None)
+    if timeout is None:
+        timeout = DEFAULT_TIMEOUT
+    kwargs["timeout"] = timeout
+
     headers = kwargs.pop("headers", None)
     if headers is None:
         headers = {}
@@ -617,7 +703,7 @@ def store_prediction(
     }
 
     ret = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/store_prediction",
+        f"{get_api_base_url_v2()}/data/store_prediction",
         method="POST",
         data=json.dumps(payload),
         headers={"Content-Type": "application/json"},
@@ -669,7 +755,7 @@ def store_predictions(
     }
 
     ret = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/store_predictions",
+        f"{get_api_base_url_v2()}/data/store_predictions",
         method="POST",
         data=json.dumps(payload),
         headers={"Content-Type": "application/json"},
@@ -725,7 +811,7 @@ def forecast_list_ref_times(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times",
+        f"{get_api_base_url_v2()}/forecast/list_ref_times",
         access_token=access_token,
         params=query,
     )
@@ -782,7 +868,7 @@ def forecast_diff(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/diff",
+        f"{get_api_base_url_v2()}/forecast/diff",
         access_token=access_token,
         params=query,
     )
@@ -790,6 +876,114 @@ def forecast_diff(
     return pd.DataFrame.from_dict(data)
 
 
+def forecast_get(
+    series_ids: List[str],
+    price_areas: List[str] = None,
+    start: str = None,
+    end: str = None,
+    *,
+    access_token: str = None,
+):
+    """Get any number of forecast time series.
+
+    This function can be used to fetch time series from the Archimedes Database.
+    To see which series are available, use `list_ids()`.
+
+    Example:
+        >>> archimedes.get(
+        >>>     series_ids=["MET/forecast_wind_speed_10m"],
+        >>>     price_areas=["NO1", "NO2"],
+        >>>     start="2022-04-20T06:00:00+00:00",
+        >>>     end="2022-04-20T12:00:00+00:00",
+        >>> )
+        series_id                 MET/forecast_wind_speed_10m
+        price_area                          NO1   NO2
+        from_dt
+        2020-06-20T04:00:00+00:00          1.30  1.30
+        2020-06-20T05:00:00+00:00          1.35  1.35
+        ...                                 ...   ...
+        2020-06-28T03:00:00+00:00          0.53  0.53
+        2020-06-28T04:00:00+00:00          0.55  0.55
+
+    Args:
+        series_ids (List[str]): The series ids to get.
+        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
+        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
+        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
+        access_token (str, optional): None - access token for the API
+
+    Returns:
+        DataFrame with all the time series data
+
+    Raises:
+        HTTPError: If an http error occurs when requesting the API.
+        NoneAuth: If the user is unauthorized or if the authorization has expired.
+    """
+
+    if isinstance(series_ids, str):
+        series_ids = [series_ids]
+
+    if isinstance(price_areas, str):
+        price_areas = [price_areas]
+
+    if start is None:
+        start = archimedes.ArchimedesConstants.DATE_LOW
+    else:
+        start = pd.to_datetime(start)
+
+    if end is None:
+        end = archimedes.ArchimedesConstants.DATE_HIGH
+    else:
+        end = pd.to_datetime(end)
+
+    queries = [
+        {
+            "start": start,
+            "end": end,
+            "flatten_columns": True,
+        }
+    ]
+    queries = [
+        {"series_ids": series_id, **query}
+        for query in queries
+        for series_id in series_ids
+    ]
+    queries = [
+        {"price_areas": price_area, **query}
+        for query in queries
+        for price_area in price_areas
+    ]
+
+    base_url = get_api_base_url_v2()
+    df = []
+    for query in queries:
+        df.extend(
+            _make_api_request(
+                f"{base_url}/forecast/get",
+                access_token=access_token,
+                params=query,
+            )
+        )
+
+    if len(df):
+        df = pd.DataFrame.from_dict(df)
+        df = df.sort_values(by=["from_dt", "ref_dt"])
+    else:
+        df = pd.DataFrame(
+            columns=["series_id", "from_dt", "ref_dt", "value", "price_area"]
+        )
+
+    df["from_dt"] = pd.to_datetime(df["from_dt"])
+    df["ref_dt"] = pd.to_datetime(df["ref_dt"])
+    df = df.pivot_table(
+        values="value",
+        columns=["series_id", "price_area"],
+        index=["from_dt", "ref_dt"],
+        aggfunc="last",
+    )
+    return df
+
+
 def forecast_get_by_ref_time_interval(
     series_id: str,
     price_area: str,
@@ -836,7 +1030,7 @@ def forecast_get_by_ref_time_interval(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval",
+        f"{get_api_base_url_v2()}/forecast/get_by_ref_time_interval",
         access_token=access_token,
         params=query,
     )
